import pytest
from iwashi.service.mirrativ import Mirrativ
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_mirrativ():
    service = Mirrativ()
    correct = Result(
        service=service,
        id="495308",
        url="https://www.mirrativ.com/user/495308",
        name="ごはんちゃんねる",
        description="YouTubeでゲーム実況やプラモ制作をしています！",
        profile_picture="https://cdn.mirrativ.com/mirrorman-prod/image/profile_image/c9780b3765e11280b17cc0a9c1f2138d32af7e27295beae3a4f1c23304d4f3e4_m.jpeg?1697778608",
        links={"https://www.youtube.com/channel/UCyA5Im3h7KNmq02ccT0KhiQ"},
    )
    await _test_service(service, correct, "https://www.mirrativ.com/user/495308")



from __future__ import annotations
import re

import bs4


from iwashi.helper import HTTP_REGEX, traverse
from iwashi.visitor import Context, Service


class Itchio(Service):
    def __init__(self) -> None:
        super().__init__(
            name="itch.io",
            regex=re.compile(
                HTTP_REGEX
                + r"((?P<id_subdomain>[\w-]+)\.itch\.io/|itch\.io/profile/(?P<id_profile>[\w-]+))",
                re.IGNORECASE,
            ),
        )

    async def resolve_id(self, context: Context, url: str) -> str | None:
        match = self.regex.match(url)
        if match is None:
            return None
        return match.group("id_profile") or match.group("id_subdomain")

    async def visit(self, context: Context, id: str):
        url = f"https://{id}.itch.io"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        user_links = soup.select(".link_group a")

        res_profile = await context.session.get(f"https://itch.io/profile/{id}")
        res_profile.raise_for_status()
        soup_profile = bs4.BeautifulSoup(await res_profile.text(), "html.parser")
        # <meta property="og:title" content="Sebastian Lague">
        name_element = soup_profile.select_one('meta[property="og:title"]')
        name = traverse(name_element).map(lambda x: x.attrs.get("content")).get()
        # <div class="avatar" style="background-image: url('https://img.itch.zone/aW1nLzEwODg5MDY5LmpwZw==/80x80%23/rdXwL%2B.jpg')"></div>
        avatar_element = soup_profile.select_one(".avatar")
        avatar = (
            traverse(avatar_element)
            .map(lambda x: x.attrs.get("style"))
            .map(str)
            .map(lambda x: x.split("url('")[1].split("')")[0])
            .get()
        )

        context.create_result(
            self,
            id=id,
            url=url,
            name=name,
            profile_picture=avatar,
        )

        for link in user_links:
            href = link.attrs.get("href")
            if href is None:
                continue
            context.enqueue_visit(href)



import pytest
from iwashi.service.instagram import Instagram
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_instagram():
    service = Instagram()
    correct = Result(
        service=service,
        id="ismsx_",
        url="https://www.instagram.com/ismsx_",
        name="いっしん",
        description="Isshin Tanaka\nmotion designer / visual artist",
        profile_picture="https://scontent-nrt1-2.cdninstagram.com/v/t51.2885-19/252132732_925276874749701_2338460024889750642_n.jpg?stp=dst-jpg_s150x150&_nc_ht=scontent-nrt1-2.cdninstagram.com&_nc_cat=109&_nc_ohc=9zhwfo0l1JUAb4kjQ9-&edm=AOQ1c0wBAAAA&ccb=7-5&oh=00_AfDFhH0vtf75VqtZ6MXSNkczWc-WMtJzVm9xxF2rIrbN8Q&oe=66293E40&_nc_sid=8b3546",
        links={"https://ismsx.jp/"},
    )
    await _test_service(service, correct)



from .bandcamp import Bandcamp
from .booth import Booth
from .fanbox import Fanbox
from .instagram import Instagram
from .linktree import Linktree
from .litlink import LitLink
from .mirrativ import Mirrativ
from .nicovideo import Nicovideo
from .note import Note
from .pixiv import Pixiv
from .reddit import Reddit
from .sketch import Sketch
from .soundcloud import Soundcloud
from .tiktok import TikTok
from .twitcasting import TwitCasting
from .twitch import Twitch
from .twitter import Twitter
from .youtube import Youtube
from .spotify import Spotify
from .github import Github
from .itchio import Itchio
from .kofi import Kofi
from .patreon import Patreon
from .skeb import Skeb


__all__ = [
    "Bandcamp",
    "Booth",
    "Fanbox",
    "Instagram",
    "Linktree",
    "LitLink",
    "Mirrativ",
    "Nicovideo",
    "Note",
    "Pixiv",
    "Reddit",
    "Sketch",
    "Soundcloud",
    "TikTok",
    "TwitCasting",
    "Twitch",
    "Twitter",
    "Youtube",
    "Spotify",
    "Github",
    "Itchio",
    "Kofi",
    "Patreon",
    "Skeb",
]

SERVICES = {
    Bandcamp(),
    Booth(),
    Fanbox(),
    Instagram(),
    Linktree(),
    LitLink(),
    Mirrativ(),
    Nicovideo(),
    Note(),
    Pixiv(),
    Reddit(),
    Sketch(),
    Soundcloud(),
    TikTok(),
    TwitCasting(),
    Twitch(),
    Twitter(),
    Youtube(),
    Spotify(),
    Github(),
    Itchio(),
    Kofi(),
    Patreon(),
    Skeb(),
}



import pytest
from iwashi.visitor import Result
from iwashi.service.fanbox import Fanbox
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_fanbox():
    service = Fanbox()
    correct = Result(
        service=service,
        id="masahiro-emotion",
        url="https://masahiro-emotion.fanbox.cc",
        name="Masahiro Emoto",
        description="English text is shown below\r\n\r\n私はアニメーション監督、キャラクターデザイナー、作画監督、イラスレーター、ロボットのビヘイビアデザインなど活動しているクリエイターです。\r\n参加作品は\r\n・攻殻機動隊\r\n・カウボーイビバップ\r\n・Animatrix\r\n・REDLINE\r\n・BLEACH\r\n\r\nなど様々なプロジェクトに参加してました。\r\n\r\nI am a creator involved in various roles such as animation director, character designer, animation supervisor, illustrator, and robot behavior design.\r\nI have contributed to projects such as Ghost in the Shell, Cowboy Bebop, Animatrix, REDLINE, BLEACH, and many others.",
        profile_picture="https://pixiv.pximg.net/c/160x160_90_a2_g5/fanbox/public/images/user/91970939/icon/aAE8bJoSKtAtHhWK17NUmMFI.jpeg",
        links={"https://x.com/masahiroemotion/"},
    )
    await _test_service(
        service,
        correct,
        "https://masahiro-emotion.fanbox.cc/",
        "https://masahiro-emotion.fanbox.cc/posts/7382756",
    )



from __future__ import annotations
import re

import bs4


from iwashi.helper import HTTP_REGEX, traverse
from iwashi.visitor import Context, Service


class Kofi(Service):
    def __init__(self) -> None:
        super().__init__(
            name="Ko-fi",
            regex=re.compile(
                HTTP_REGEX + r"ko-fi\.com/(?P<id>[\w-]+)",
                re.IGNORECASE,
            ),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://ko-fi.com/{id}"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        name = (
            traverse(soup.select_one(".mob-profile-name"))
            .map(lambda x: x.text)
            .map(str.strip)
            .get()
        )
        avatar = (
            traverse(soup.select_one("#profilePicture2"))
            .map(lambda x: x.attrs.get("src"))
            .map(str)
            .get()
        )
        social_links = soup.select(".social-link a")

        context.create_result(
            self,
            id=id,
            url=url,
            name=name,
            profile_picture=avatar,
        )

        for link in social_links:
            href = link.attrs.get("href")
            if href is None:
                continue
            context.enqueue_visit(href)



import pytest
from iwashi.service.reddit import Reddit
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_reddit():
    service = Reddit()
    correct = Result(
        service=service,
        id="MarcusFaze",
        url="https://www.reddit.com/user/MarcusFaze",
        name="MarcusFaze",
        description="I’m 18, I’m a Wrestling Fan,I’m from North Carolina",
        profile_picture="https://styles.redditmedia.com/t5_25b6em/styles/profileIcon_ym46r18i589b1.jpg?width=256&amp;height=256&amp;crop=256:256,smart&amp;s=7b7e1b4c208f847f36de445ebd5b8f0c32657071",
        links={"https://twitter.com/marcus_faze"},
    )
    await _test_service(service, correct, "https://www.reddit.com/user/MarcusFaze")



from typing import TypedDict


class webCommandMetadata(TypedDict):
    apiUrl: str


class commandMetadata(TypedDict):
    webCommandMetadata: webCommandMetadata


class continuationCommand(TypedDict):
    token: str


class continuationEndpoint(TypedDict):
    commandMetadata: commandMetadata
    continuationCommand: continuationCommand


class continuationItemRenderer(TypedDict):
    continuationEndpoint: continuationEndpoint


class itemSectionRenderercontentsItem(TypedDict):
    continuationItemRenderer: continuationItemRenderer


type itemSectionRenderercontents = list[itemSectionRenderercontentsItem]


class itemSectionRenderer(TypedDict):
    contents: itemSectionRenderercontents


class contentsItem(TypedDict):
    itemSectionRenderer: itemSectionRenderer


type contents = list[contentsItem]


class sectionListRenderer(TypedDict):
    contents: contents


class content(TypedDict):
    sectionListRenderer: sectionListRenderer


class engagementPanelSectionListRenderer(TypedDict):
    content: content


class engagementPanel(TypedDict):
    engagementPanelSectionListRenderer: engagementPanelSectionListRenderer


class showEngagementPanelEndpoint(TypedDict):
    engagementPanel: engagementPanel


class innertubeCommand(TypedDict):
    showEngagementPanelEndpoint: showEngagementPanelEndpoint


class onTap(TypedDict):
    innertubeCommand: innertubeCommand


class commandRunsItem(TypedDict):
    onTap: onTap


type commandRuns = list[commandRunsItem]


class more(TypedDict):
    commandRuns: commandRuns


class channelHeaderLinksViewModel(TypedDict):
    more: more


class headerLinks(TypedDict):
    channelHeaderLinksViewModel: channelHeaderLinksViewModel


class c4TabbedHeaderRenderer(TypedDict):
    headerLinks: headerLinks


class header(TypedDict):
    c4TabbedHeaderRenderer: c4TabbedHeaderRenderer


class ProfileRes3(TypedDict):
    header: header


# data: ProfileRes2 = ...

# for a in data['header']['c4TabbedHeaderRenderer']['headerLinks']['channelHeaderLinksViewModel']['more']['commandRuns']:
#     for b in a['onTap']['innertubeCommand']['showEngagementPanelEndpoint']['engagementPanel']['engagementPanelSectionListRenderer']['content']['sectionListRenderer']['contents']:
#         for c in b['itemSectionRenderer']['contents']:
#             endpoint = c['continuationItemRenderer']['continuationEndpoint']
#             if endpoint['commandMetadata']['webCommandMetadata']['apiUrl'].startswith('/youtubei/v1/browse'):
#                 return endpoint['continuationCommand']['token']



import re
from typing import List, TypedDict

from loguru import logger

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class Fanbox(Service):
    def __init__(self) -> None:
        super().__init__(
            name="Fanbox",
            regex=re.compile(HTTP_REGEX + r"(?P<id>[\w-]+)\.fanbox\.cc", re.IGNORECASE),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://{id}.fanbox.cc"
        creator_res = await context.session.get(
            f"https://api.fanbox.cc/creator.get?creatorId={id}",
            headers={
                "accept": "application/json",
                "origin": f"https://{id}.fanbox.cc",
                "referer": f"https://{id}.fanbox.cc/",
            },
        )
        creator_res.raise_for_status()
        if creator_res.status // 100 == 4:
            logger.warning(f"[Fanbox] Could not find user for {url}")
            return

        info: Root = await creator_res.json()
        if "error" in info:
            logger.warning(f"[Fanbox] Could not find user for {url}")
            return
        context.create_result(
            self,
            id=id,
            url=url,
            name=info["body"]["user"]["name"],
            description=info["body"]["description"],
            profile_picture=info["body"]["user"]["iconUrl"],
        )

        for link in info["body"]["profileLinks"]:
            context.enqueue_visit(link)


class User(TypedDict):
    userId: str
    name: str
    iconUrl: str


class ProfileItemsItem0(TypedDict):
    id: str
    type: str
    imageUrl: str
    thumbnailUrl: str


class Body(TypedDict):
    user: User
    creatorId: str
    description: str
    hasAdultContent: bool
    coverImageUrl: str
    profileLinks: List[str]
    profileItems: List[ProfileItemsItem0]
    isFollowed: bool
    isSupported: bool
    isStopped: bool
    isAcceptingRequest: bool
    hasBoothShop: bool


class Root(TypedDict):
    body: Body



import asyncio
from typing import List, MutableSet, NamedTuple, Optional
import aiohttp
from loguru import logger
from .helper import BASE_HEADERS, DEBUG, normalize_url
from .visitor import Context, Result, Service, Visitor
from .service import SERVICES


class Identifier(NamedTuple):
    site: str
    id: str


class Iwashi(Visitor):
    def __init__(self) -> None:
        self.services: List[Service] = []
        self.visited_urls: MutableSet[str] = set()
        self.visited_ids: MutableSet[Identifier] = set()
        self.tasks: List[asyncio.Task] = []
        self.session = aiohttp.ClientSession(headers=BASE_HEADERS)

    def add_service(self, service: Service) -> None:
        self.services.append(service)

    def is_visited(self, url: str) -> bool:
        return url in self.visited_urls

    def mark_visited(self, url: str):
        self.visited_urls.add(url)

    async def tree(self, url: str, context: Optional[Context] = None) -> Result | None:
        async with self.session:
            context = context or Context(session=self.session, visitor=self)
            context = context.create_context()
            result = await self.visit(url, context)
            while self.tasks:
                await self.tasks.pop()

        return result

    def enqueue_visit(self, url: str, context: Context) -> None:
        coro = self.visit(url, context)
        task = asyncio.create_task(coro)
        self.tasks.append(task)

    async def visit(self, url: str, context: Context) -> Optional[Result]:
        normalized_url = normalize_url(url)
        if normalized_url is None:
            return None
        context = context.create_context()
        if self.is_visited(normalized_url):
            return None
        for service in self.services:
            match = service.match(normalized_url, context)
            if match is None:
                continue

            try:
                id = await service.resolve_id(context, normalized_url)
                if id is None:
                    continue
                identifier = Identifier(site=service.name, id=id)
                if identifier in self.visited_ids:
                    continue
                self.visited_ids.add(identifier)
                async with service.throttle:
                    await service.visit(context, id)
            except Exception as e:
                logger.warning(
                    f"[Service Error] {service.name} failed to visit {normalized_url}"
                )
                logger.exception(e)
                if DEBUG:
                    raise e
                continue
            self.mark_visited(normalized_url)
            break
        else:
            self.mark_visited(normalized_url)
            if await self.try_redirect(normalized_url, context):
                return context.result
            else:
                logger.warning(f"[No Service] No service matched {normalized_url}")

        return context.result

    async def try_redirect(self, url: str, context: Context) -> bool:
        try:
            res = await context.session.get(
                url,
                headers=BASE_HEADERS
                | {
                    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 Edg/124.0.0.0",
                    "Referer": url,
                },
                allow_redirects=True,
                timeout=5,
            )
            res.raise_for_status()
        except aiohttp.ClientError as e:
            logger.exception(e)
            logger.warning(f"[Redirect] failed to redirect {url}")
            return False
        except asyncio.TimeoutError as e:
            logger.exception(e)
            logger.warning(f"[Redirect] failed to redirect {url}")
            return False
        new_url = str(res.url)
        if new_url == url:
            return False
        context.enqueue_visit(new_url)
        logger.info(f"[Redirect] {url} -> {new_url}")
        return True


def get_iwashi():
    iwashi = Iwashi()

    for service in SERVICES:
        iwashi.add_service(service)

    return iwashi


async def tree(url: str, iwashi: Optional[Iwashi] = None) -> Optional[Result]:
    iwashi = iwashi or get_iwashi()
    return await iwashi.tree(url)



import pytest
from iwashi.visitor import Result
from iwashi.service.booth import Booth
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_booth():
    service = Booth()
    correct = Result(
        service=service,
        id="miluda",
        url="https://miluda.booth.pm",
        name="Sunshine Mill 太陽光工場",
        description="Welcome to 太陽光工場.\r\nMiludaと申します。同人初心者アメリカ人ですけどどうぞよろしくお願いします。メインは東方や艦これです。\r\n\r\nよろしくお願いします！",
        profile_picture="https://booth.pximg.net/c/128x128/users/274/icon_image/41f84dbd-19af-49b9-8cf5-888eb389e500_base_resized.jpg",
        links={
            "https://miluda.com",
            "https://twitter.com/Miluda",
            "https://www.pixiv.net/users/161480",
        },
    )
    await _test_service(
        service,
        correct,
        "https://miluda.booth.pm/",
        "https://miluda.booth.pm/items/397",
    )



import pytest
from iwashi.service.tiktok import TikTok
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_tiktok():
    service = TikTok()
    correct = Result(
        service=service,
        id="cocoyola",
        url="https://www.tiktok.com/@cocoyola",
        name="☀️🥂🫶🏻⭐️🍸",
        description="-everything you’re looking for and more-\n\n🪸🌅🥂🫶🏻🍸",
        profile_picture=".tiktokcdn.com/tos-maliva-avt-0068/7325124622654341126~c5_1080x1080.jpeg",
        links=set(),
    )
    await _test_service(
        service,
        correct,
        "https://www.tiktok.com/@cocoyola?lang=ja-JP",
        "https://www.tiktok.com/@cocoyola/video/7234867719043157274?lang=ja-JP",
    )



from __future__ import annotations

import re
from typing import Dict, List, TypedDict, Union

import bs4

from iwashi.helper import BASE_HEADERS, HTTP_REGEX
from iwashi.visitor import Context, Service


class Mirrativ(Service):
    def __init__(self):
        super().__init__(
            name="Mirrativ",
            regex=re.compile(
                HTTP_REGEX + r"mirrativ.com/user/(?P<id>[\d]+)", re.IGNORECASE
            ),
        )

    async def fetch_scrf_token(self, context: Context) -> str | None:
        res = await context.session.get("https://www.mirrativ.com/")
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        element = soup.select_one('meta[name="csrf-token"]')
        if element is None:
            raise RuntimeError("Could not find csrf-token")
        return element.attrs.get("content")

    async def visit(self, context: Context, id: str):
        url = f"https://www.mirrativ.com/user/{id}"

        headers = BASE_HEADERS | {
            "accept": "application/json",
            "x-csrf-token": await self.fetch_scrf_token(context),
        }

        res = await context.session.get(
            "https://www.mirrativ.com/api/user/profile",
            params={
                "user_id": id,
            },
            headers=headers,
        )
        res.raise_for_status()
        info: Root = await res.json()
        context.create_result(
            self,
            id=id,
            url=url,
            name=info["name"],
            description=info["description"],
            profile_picture=info["profile_image_url"],
        )

        for link in info["links"]:
            context.enqueue_visit(link["url"])


class PreviewViewersItem(TypedDict):
    profile_frame_image_url: str
    badge_image_url: str
    profile_image_url: str
    yell_rank: int
    user_id: str


class SeasonYell(TypedDict):
    is_display_target: bool
    preview_viewers: List[PreviewViewersItem]
    total_viewer_count: int


class BadgesItem(TypedDict):
    image_url: str
    small_image_url: str


class LinksItem(TypedDict):
    url: str


class MutualFollowees(TypedDict):
    text: str
    image_urls: List


class Onlive(TypedDict):
    live_id: str


class Status(TypedDict):
    msg: str
    ok: int
    error: str
    captcha_url: str
    error_code: int
    message: str


class RibbonsItem(TypedDict):
    is_continuous_ribbon: int
    label_remaining_period: None
    image_url: str
    ribbon_id: int
    is_label: int


class RibbonsItem0(TypedDict):
    is_continuous_ribbon: int
    image_url: str
    is_label: int


class LabelRemainingPeriod(TypedDict):
    remain_seconds: int
    text: str
    is_highlight: int


class RibbonsItem1(TypedDict):
    is_continuous_ribbon: int
    label_remaining_period: LabelRemainingPeriod
    image_url: str
    ribbon_id: int
    is_label: int


class SeasonRating(TypedDict):
    class_name: str
    icon_url: str


class Root(TypedDict):
    avatar_body_image_url: str
    ambassador_image_url: str
    current_continuous_record: str
    registered_at: str
    custom_thanks_message: str
    profile_image_url: str
    season_yell: SeasonYell
    badges: List[BadgesItem]
    continuous_type: int
    is_visible_birthday: bool
    follower_num: str
    next_continuous_streamer_badge_url: str
    anniversary: Dict
    my_app_num: int
    links: List[LinksItem]
    mutual_followees: MutualFollowees
    grade_id: int
    twitter_screen_name: str
    birthday_to: int
    name: str
    description: str
    birthday_editable_date: str
    is_birthday: int
    properties: List
    total_viewer_num: int
    profile_frame_image_url: str
    live_announcement: None
    is_blocking: int
    is_blocked: int
    user_id: str
    paypal_username: str
    next_continuous_streamer_text: str
    onlive: Onlive
    share_url: str
    is_able_continuous_stream_holiday: int
    status: Status
    ribbons: List[Union[RibbonsItem, RibbonsItem0, RibbonsItem1]]
    birthday_from: int
    tutorial_mission: None
    is_birthday_editable: int
    avatar_background_image_url: str
    is_new: int
    following_num: str
    is_cheerleader: int
    catalog_label_image_url: str
    kakao_name: str
    is_follower: int
    has_started_first_live: int
    max_continuous_record: str
    season_rating: SeasonRating
    birthday: str
    user_level: Dict
    remaining_days_for_continuous_streamer: int
    continuous_achieved_text: str
    chat_enabled: int
    is_continuous_streamer: int
    is_following: int
    live_request_num: str



from __future__ import annotations

import json
import re
from typing import List, TypedDict

import bs4

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class Note(Service):
    def __init__(self):
        super().__init__(
            name="Note",
            regex=re.compile(HTTP_REGEX + r"note\.com/(?P<id>[^/]+)", re.IGNORECASE),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://note.com/{id}"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")

        data_element = soup.select_one("script[type='application/ld+json']")
        if data_element is None:
            return
        data_root: Root = json.loads(data_element.text)
        if len(data_root) != 1:
            raise Exception("Root element is not unique")
        data = data_root[0]["mainEntity"]

        context.create_result(
            self,
            id=id,
            url=url,
            name=data["name"],
            description=data["description"],
            profile_picture=data["image"]["url"],
        )

        links: set[str] = set()
        for element in soup.select(".m-creatorSocialLinks__item"):
            link = element.select_one("a")
            if link is None:
                continue
            if "href" not in link.attrs:
                continue
            links.add(link.attrs["href"])
        for link in links:
            context.enqueue_visit(link)


Image = TypedDict(
    "Image", {"@type": "str", "url": "str", "width": "int", "height": "int"}
)
Mainentity = TypedDict(
    "Mainentity",
    {
        "@type": "str",
        "name": "str",
        "url": "str",
        "image": "Image",
        "description": "str",
    },
)
RootItem = TypedDict(
    "RootItem",
    {
        "@context": "str",
        "@type": "str",
        "dateCreated": "str",
        "mainEntity": "Mainentity",
    },
)
Root = List[RootItem]



import json
import re
from typing import Any
from urllib import parse

import bs4

from iwashi.helper import HTTP_REGEX, normalize_url
from iwashi.service.youtube.types.ytinitialdata2 import ProfileRes2
from iwashi.service.youtube.types.ytinitialdata3 import ProfileRes3
from iwashi.visitor import Context, Service

from .types import thumbnails, ytinitialdata
from .types.about import AboutRes

VANITY_ID_REGEX = re.compile(r"youtube.com/@(?P<id>[\w\-.]+)")


class Youtube(Service):
    def __init__(self):
        super().__init__(
            name="Youtube",
            regex=re.compile(
                HTTP_REGEX + r"((m|gaming)\.)?(youtube\.com|youtu\.be)",
            ),
        )

    async def resolve_id(self, context: Context, url: str) -> str | None:
        normalized_url = normalize_url(url)
        if normalized_url is None:
            return None
        uri = parse.urlparse(normalized_url)
        if uri.hostname == "youtu.be":
            return await self._channel_by_video(context, uri.path[1:])
        type = next(filter(None, uri.path.split("/")))
        if type.startswith("@"):
            return self._id_from_vanity_url(url)
        if type == "playlist":
            return None
        if type == "watch":
            return await self._channel_by_video(
                context, parse.parse_qs(uri.query)["v"][0]
            )
        if type in ("channel", "user", "c"):
            return await self._channel_by_url(context, url)
        if len(uri.path) > 1:
            maybe_vanity = uri.path.split("/")[1]
            return self._id_from_vanity_url(f"https://youtube.com/@{maybe_vanity}")
        return None

    async def _channel_by_video(self, context: Context, video_id: str) -> str | None:
        result = await self._channel_by_oembed(context, video_id)
        if result is not None:
            return result
        response = await context.session.get(
            f"https://www.youtube.com/watch?v={video_id}"
        )
        response.raise_for_status()
        soup = bs4.BeautifulSoup(await response.text(), "html.parser")
        element = soup.select_one('span[itemprop="author"] > link[itemprop="url"]')
        if element is None:
            return None
        href = element.attrs.get("href")
        if href is None:
            return None
        return self._id_from_vanity_url(href)

    async def _channel_by_oembed(self, context: Context, video_id: str) -> str | None:
        res = await context.session.get(
            "https://www.youtube.com/oembed",
            params={
                "url": f"https://www.youtube.com/watch?v={video_id}",
                "format": "json",
            },
        )
        res.raise_for_status()
        data = await res.json()
        author_url = data.get("author_url")
        if author_url is None:
            return None
        return self._id_from_vanity_url(author_url)

    async def _channel_by_url(self, context: Context, url: str) -> str | None:
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        data = self.extract_initial_data(soup)
        vanity_url = data["metadata"]["channelMetadataRenderer"]["vanityChannelUrl"]
        return self._id_from_vanity_url(vanity_url)

    def _id_from_vanity_url(self, url: str) -> str | None:
        match = VANITY_ID_REGEX.search(url)
        if match is None:
            return None
        return match.group("id")

    def parse_thumbnail(self, thumbnails: thumbnails) -> str:
        size = 0
        url: str | None = None
        for thumbnail in thumbnails["thumbnails"]:
            if thumbnail["width"] > size:
                size = thumbnail["width"]
                url = thumbnail["url"]
        if url is None:
            raise RuntimeError("Thumbnail not found")
        return url

    async def get_token(self, data: Any) -> str | None:
        data2: ProfileRes2 = data
        if "pageHeaderRenderer" in data2["header"]:
            suffix = data2["header"]["pageHeaderRenderer"]["content"][
                "pageHeaderViewModel"
            ]["attribution"]["attributionViewModel"]["suffix"]
            if not suffix:
                return None
            for a in suffix["commandRuns"]:
                for b in a["onTap"]["innertubeCommand"]["showEngagementPanelEndpoint"][
                    "engagementPanel"
                ]["engagementPanelSectionListRenderer"]["content"][
                    "sectionListRenderer"
                ]["contents"]:
                    for c in b["itemSectionRenderer"]["contents"]:
                        endpoint = c["continuationItemRenderer"]["continuationEndpoint"]
                        if endpoint["commandMetadata"]["webCommandMetadata"][
                            "apiUrl"
                        ].startswith("/youtubei/v1/browse"):
                            return endpoint["continuationCommand"]["token"]
        else:
            data3: ProfileRes3 = data
            for a in data3["header"]["c4TabbedHeaderRenderer"]["headerLinks"][
                "channelHeaderLinksViewModel"
            ]["more"]["commandRuns"]:
                for b in a["onTap"]["innertubeCommand"]["showEngagementPanelEndpoint"][
                    "engagementPanel"
                ]["engagementPanelSectionListRenderer"]["content"][
                    "sectionListRenderer"
                ]["contents"]:
                    for c in b["itemSectionRenderer"]["contents"]:
                        endpoint = c["continuationItemRenderer"]["continuationEndpoint"]
                        if endpoint["commandMetadata"]["webCommandMetadata"][
                            "apiUrl"
                        ].startswith("/youtubei/v1/browse"):
                            return endpoint["continuationCommand"]["token"]
        return None

    def parse_redirect(self, url: str) -> str:
        uri = parse.urlparse(url)
        if uri.hostname != "www.youtube.com":
            return url
        if uri.path == "/redirect":
            return parse.parse_qs(uri.query)["q"][0]
        return url

    async def visit(self, context: Context, id: str):
        url = f"https://www.youtube.com/@{id}"
        res = await context.session.get(url)

        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        data = self.extract_initial_data(soup)
        vanity_id = data["metadata"]["channelMetadataRenderer"]["vanityChannelUrl"]
        name = data["metadata"]["channelMetadataRenderer"]["title"]
        description = data["metadata"]["channelMetadataRenderer"]["description"]
        profile_picture = self.parse_thumbnail(
            data["metadata"]["channelMetadataRenderer"]["avatar"]
        )
        context.create_result(
            self,
            id=id,
            url=f"https://www.youtube.com/@{vanity_id.split('@')[1]}",
            name=name,
            description=description,
            profile_picture=profile_picture,
        )
        token = await self.get_token(data)
        if token is None:
            return
        about_res = await context.session.post(
            "https://www.youtube.com/youtubei/v1/browse",
            json={
                "context": {
                    "client": {
                        "userAgent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36,gzip(gfe)",
                        "clientName": "WEB",
                        "clientVersion": "2.20240509.00.00",
                    },
                },
                "continuation": token,
            },
        )
        about_res.raise_for_status()
        about_data: AboutRes = await about_res.json()
        links: list[str] = []
        for a in about_data["onResponseReceivedEndpoints"]:
            for b in a["appendContinuationItemsAction"]["continuationItems"]:
                for c in b["aboutChannelRenderer"]["metadata"]["aboutChannelViewModel"][
                    "links"
                ]:
                    for d in c["channelExternalLinkViewModel"]["link"]["commandRuns"]:
                        url = d["onTap"]["innertubeCommand"]["urlEndpoint"]["url"]
                        links.append(url)

        for link in links:
            context.enqueue_visit(self.parse_redirect(link))

    def extract_initial_data(self, soup: bs4.BeautifulSoup) -> ytinitialdata:
        for script in soup.select("script"):
            if script.string is None:
                continue
            match = re.search(r"ytInitialData\s*=\s*({.*?});", script.string)
            if match is not None:
                data: ytinitialdata = json.loads(match.group(1))
                break
        else:
            raise RuntimeError("ytInitialData not found")
        return data



import pytest
from iwashi.service.patreon import Patreon
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_itchio():
    service = Patreon()
    correct = Result(
        service=service,
        id="sebastianlague",
        url="https://www.patreon.com/sebastianlague",
        name="Sebastian Lague",
        description="Hello! I create game development tutorials on a wide range of topics. These videos are all freely available on my youtube channel.If you'd like to support the creation of more content like this, you can set a monthly recurring pledge here. There are some small perks on offer in return for your generosity –– see the sidebar for more info.",
        profile_picture="https://c10.patreonusercontent.com/4/patreon-media/p/campaign/114575/0a25704faa914c03839a205f218d36f6/eyJ3Ijo2MjB9/1.jpg",
        links={
            "https://twitter.com/SebastianLague",
            "https://youtube.com/channel/UCmtyQOKKmrMVaKuRXz02jbQ",
        },
    )
    await _test_service(
        service,
        correct,
        "https://www.patreon.com/sebastianlague",
        "https://www.patreon.com/sebastianlague/about",
    )



import pytest
from iwashi.service.github import Github
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_github():
    service = Github()
    correct = Result(
        service=service,
        id="astral-sh",
        url="https://github.com/astral-sh",
        name="Astral",
        description="High-performance developer tools for the Python ecosystem.",
        profile_picture="https://avatars.githubusercontent.com/u/115962839",
        links={
            "https://twitter.com/astral_sh",
        },
    )
    await _test_service(
        service,
        correct,
        "https://github.com/astral-sh",
        "https://github.com/astral-sh/rye",
    )



from typing import TypedDict


class WebCommandMetadata(TypedDict):
    apiUrl: str


class CommandMetadata(TypedDict):
    webCommandMetadata: WebCommandMetadata


class ContinuationCommand(TypedDict):
    token: str


class ContinuationEndpoint(TypedDict):
    commandMetadata: CommandMetadata
    continuationCommand: ContinuationCommand


class ContinuationItemRenderer(TypedDict):
    continuationEndpoint: ContinuationEndpoint


class ItemSectionRendererContentsItem(TypedDict):
    continuationItemRenderer: ContinuationItemRenderer


type ItemSectionRendererContents = list[ItemSectionRendererContentsItem]


class ItemSectionRenderer(TypedDict):
    contents: ItemSectionRendererContents


class SectionListRendererContentsItem(TypedDict):
    itemSectionRenderer: ItemSectionRenderer


type SectionListRendererContents = list[SectionListRendererContentsItem]


class SectionListRenderer(TypedDict):
    contents: SectionListRendererContents


class EngagementPanelSectionListRendererContent(TypedDict):
    sectionListRenderer: SectionListRenderer


class EngagementPanelSectionListRenderer(TypedDict):
    content: EngagementPanelSectionListRendererContent


class EngagementPanel(TypedDict):
    engagementPanelSectionListRenderer: EngagementPanelSectionListRenderer


class ShowEngagementPanelEndpoint(TypedDict):
    engagementPanel: EngagementPanel


class InnertubeCommand(TypedDict):
    showEngagementPanelEndpoint: ShowEngagementPanelEndpoint


class OnTap(TypedDict):
    innertubeCommand: InnertubeCommand


class CommandRunsItem(TypedDict):
    onTap: OnTap


type CommandRuns = list[CommandRunsItem]


class Suffix(TypedDict):
    commandRuns: CommandRuns


class AttributionViewModel(TypedDict):
    suffix: Suffix


class Attribution(TypedDict):
    attributionViewModel: AttributionViewModel


class PageHeaderViewModel(TypedDict):
    attribution: Attribution


class Content(TypedDict):
    pageHeaderViewModel: PageHeaderViewModel


class PageHeaderRenderer(TypedDict):
    content: Content


class Header(TypedDict):
    pageHeaderRenderer: PageHeaderRenderer


class ProfileRes2(TypedDict):
    header: Header


# data: ProfileRes = ...

# token: str | None = None
# for a in data['header']['pageHeaderRenderer']['content']['pageHeaderViewModel']['attribution']['attributionViewModel']['suffix']['commandRuns']:
#     for b in a['onTap']['innertubeCommand']['showEngagementPanelEndpoint']['engagementPanel']['engagementPanelSectionListRenderer']['content']['sectionListRenderer']['contents']:
#         for c in b['itemSectionRenderer']['contents']:
#             endpoint = c['continuationItemRenderer']['continuationEndpoint']
#             if endpoint['commandMetadata']['webCommandMetadata']['apiUrl'].startswith('/youtubei/v1/browse'):
#                 token = endpoint['continuationCommand']['token']
#                 break



from __future__ import annotations

import json
import re
from typing import Dict, List, TypedDict, Union

import bs4
from loguru import logger

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class LitLink(Service):
    def __init__(self) -> None:
        super().__init__(
            name="LitLink",
            regex=re.compile(HTTP_REGEX + r"lit\.link/(?P<id>\w+)", re.IGNORECASE),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://lit.link/{id}"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        data_element = soup.find(attrs={"id": "__NEXT_DATA__"})
        if data_element is None:
            logger.warning(f"[LitLink] Could not find data element for {url}")
            return
        data: Root = json.loads(data_element.get_text())
        profile = data["props"]["pageProps"]["profile"]

        context.create_result(
            self,
            id=id,
            url=url,
            name=profile["name"],
            description=profile["profileText"],
            profile_picture=profile["pictureUrl"],
        )

        for link in profile["snsIconLinks"]:
            if "url" not in link:
                continue
            context.enqueue_visit(link["url"])
        for link in profile["profileLinks"]:
            if "buttonLink" not in link:
                continue
            button_link = link["buttonLink"]
            if button_link is None:
                continue
            context.enqueue_visit(button_link["url"])


class SnsiconlinksItem(TypedDict):
    id: str
    type: str
    url: str
    userId: str
    creatorId: str


class Textlink(TypedDict):
    id: str
    title: None
    description: str
    publicationStartAt: None
    publicationEndAt: None
    userId: str
    creatorId: str


class ProfilelinksItem(TypedDict):
    profileLinkType: str
    buttonLink: None
    textLink: Textlink
    imageLink: None
    movieLink: None
    musicLink: None
    shopLink: None
    marginBlock: None


class Buttonlink(TypedDict):
    id: str
    iconUrl: str
    title: str
    description: None
    url: str
    urlType: str
    publicationStartAt: None
    publicationEndAt: None
    userId: str
    creatorId: str


class ProfilelinksItem0(TypedDict):
    profileLinkType: str
    buttonLink: Buttonlink
    textLink: None
    imageLink: None
    movieLink: None
    musicLink: None
    shopLink: None
    marginBlock: None


class Textlink0(TypedDict):
    id: str
    title: str
    description: str
    publicationStartAt: None
    publicationEndAt: None
    userId: str
    creatorId: str


class ProfilelinksItem1(TypedDict):
    profileLinkType: str
    buttonLink: None
    textLink: Textlink0
    imageLink: None
    movieLink: None
    musicLink: None
    shopLink: None
    marginBlock: None


class ProfileimagesItem(TypedDict):
    id: str
    iconUrl: None
    imageUrl: str
    title: None
    description: None
    url: str
    urlType: str
    imageLinkId: str


class Imagelink(TypedDict):
    id: str
    type: str
    profileImages: List[ProfileimagesItem]
    publicationStartAt: None
    publicationEndAt: None
    userId: str
    creatorId: str


class ProfilelinksItem2(TypedDict):
    profileLinkType: str
    buttonLink: None
    textLink: None
    imageLink: Imagelink
    movieLink: None
    musicLink: None
    shopLink: None
    marginBlock: None


class ProfileimagesItem0(TypedDict):
    id: str
    iconUrl: None
    imageUrl: str
    title: None
    description: None
    url: None
    urlType: str
    imageLinkId: str


class Imagelink0(TypedDict):
    id: str
    type: str
    profileImages: List[ProfileimagesItem0]
    publicationStartAt: None
    publicationEndAt: None
    userId: str
    creatorId: str


class ProfilelinksItem3(TypedDict):
    profileLinkType: str
    buttonLink: None
    textLink: None
    imageLink: Imagelink0
    movieLink: None
    musicLink: None
    shopLink: None
    marginBlock: None


class Creatordetaillayout(TypedDict):
    id: str
    fontFamily: str
    fontColor: str
    fontSize: str
    textAlign: str
    backgroundImageUrl: str
    backgroundColor: str
    backgroundGradation: str
    backgroundOverlayColor: str
    linkShape: str
    linkColor: str
    template: str
    userId: str
    creatorId: str


class Profile(TypedDict):
    uid: str
    userId: str
    creatorId: str
    name: str
    catchphrase: str
    sex: str
    birthday: str
    profileText: None
    urlPath: str
    pictureUrl: str
    pictureType: str
    snsIconLinks: List[SnsiconlinksItem]
    profileLinks: List[
        Union[
            ProfilelinksItem1,
            ProfilelinksItem0,
            ProfilelinksItem,
            ProfilelinksItem2,
            ProfilelinksItem3,
        ]
    ]
    creatorDetailLayout: Creatordetaillayout
    creatorSnsActivityCategories: List


class Account(TypedDict):
    email: str
    url: str
    emailUpdateResponse: None
    isLoading: bool
    lineConnected: bool
    emailConnected: bool
    urlUpdateResponse: None
    urlUpdateError: None


class Creatordetailedit(TypedDict):
    error: None
    isEdit: bool
    editingProfile: None
    isLoading: bool
    imageUpLoading: bool
    showSavedToast: bool
    toastText: str
    selectedIndexOnImageOrSnsModal: int
    showIconQrCode: bool
    hasSavedProfile: bool


class Genrecategory(TypedDict):
    isLoading: bool
    selectedMoreThanOne: bool
    selectedSnsActivityCategoryIds: List
    selectedSnsActivityGenreIds: List
    apiError: None


class Profile0(TypedDict):
    isLoggedIn: bool
    showCopiedMessage: bool
    showIconQrCode: bool
    profile: None
    isSendViewTypeAccessLog: bool


class Linelogin(TypedDict):
    isLoading: bool
    apiError: None
    lineLoginResponse: None
    hasInit: bool
    isError: bool


class Login(TypedDict):
    apiError: None
    loginResponse: None
    isLoading: bool
    loginErrorMessageId: None


class Modal(TypedDict):
    modalComponent: None
    modalOpened: bool
    masterModalId: str
    modalMaxWidth: int
    isFullSizeInMobile: bool
    onCloseModal: None


class Linemessaging(TypedDict):
    lineMessaging: None
    isLoading: bool


class Passwordreminder(TypedDict):
    passwordReminderResponse: None
    isLoading: bool
    isCompletedSendEmail: bool
    hasErrorResponse: bool


class Signup(TypedDict):
    isSignUpSuccess: bool
    signUpByLineResponse: None
    signUpEmailResponse: None
    isLoading: bool
    registeredAlready: bool
    defaultEmail: str
    hasErrorSignupResponse: bool


class Firebaseauth(TypedDict):
    firebaseUser: None
    isAuthLoading: bool
    isResendEmailVerificationSucceeded: None


class Signupdetail(TypedDict):
    isInstagramConnected: bool
    isTwitterConnected: bool
    isLoading: bool
    isVerifiedUrl: None


class Linesignup(TypedDict):
    apiError: bool
    lineSignUpResponse: None
    isLoading: bool
    signUpApiResponse: None


class DatasetsItem(TypedDict):
    label: str
    backgroundColor: str
    borderColor: str
    fill: bool
    tension: float
    data: List


class Usergraphaccesslog(TypedDict):
    labels: List
    datasets: List[DatasetsItem]


class Analytics(TypedDict):
    displayPeriod: str
    urlSortType: str
    topSortType: str
    isUrlSortAscendant: bool
    isTopSortAscendant: bool
    isReferralSortAscendant: bool
    isDeviceSortAscendant: bool
    pvCounts: int
    clickCounts: int
    accessTopTableSortType: str
    userTodayAccessLog: None
    userOneWeekAccessLog: None
    userOneMonthAccessLog: None
    userThreeMonthsAccessLog: None
    userSixMonthsAccessLog: None
    userOneYearAccessLog: None
    userAllAccessLog: None
    userGraphAccessLog: Usergraphaccesslog
    userUrlAccessLogs: List
    userTopAccessLogs: List
    userReferralAccessLogs: List
    userDeviceAccessLogs: List
    isAnalyticsStateLoading: bool
    urlAddedAreaHeight: int
    topAddedAreaHeight: int
    referralAddedAreaHeight: int
    isShowingMoreOnUrl: bool
    isShowingMoreOnTop: bool
    isShowingMoreOnReferral: bool
    isAnalyticsApiError: bool
    isShowingToast: bool
    apiError: None


class Notificationlist(TypedDict):
    isLoading: bool
    selectedNotificationId: None


class Creatordetailedittutorial(TypedDict):
    editingCreatorPreference: None
    tutorialCount: int
    isTutorialButtonEditDone: bool
    isTutorialLinkDraggerDone: bool
    isTutorialLinkEditDone: bool


class Accountdelete(TypedDict):
    isLoading: bool


class Profileimagenftmodal(TypedDict):
    isLoading: bool
    error: None


class Signupgenre(TypedDict):
    isSucceeded: None
    isLoading: bool
    selectedMoreThanOne: bool
    selectedSnsActivityCategoryIds: List
    selectedSnsActivityGenreIds: List


class Maintenance(TypedDict):
    isLoading: bool
    litlinkError: None
    isFirestoreMaintenanceMode: bool
    isApiMaintenanceMode: bool


class Authentication(TypedDict):
    userId: str
    accessToken: str
    apiError: None


class Urlalertmodal(TypedDict):
    alertUrlType: None
    alertMovieType: None
    alertMusicType: None
    alertShopType: None


class Profilesnsmodal(TypedDict):
    error: None
    editingSnsIconLinkDetails: List
    modalSnsType: None
    snsModalDefaultUrl: str
    updatingSnsIconLinkId: None


class Creator(TypedDict):
    creators: Dict
    currentCreatorId: None


class Creatordetaillayout0(TypedDict):
    creatorDetailLayouts: Dict


class Creatorsnsactivitycategory(TypedDict):
    creatorSnsActivityCategories: Dict


class Snsiconlink(TypedDict):
    snsIconLinks: Dict


class User(TypedDict):
    users: Dict


class Buttonlink0(TypedDict):
    buttonLinks: Dict


class Textlink1(TypedDict):
    textLinks: Dict


class Imagelink1(TypedDict):
    imageLinks: Dict


class Movielink(TypedDict):
    movieLinks: Dict


class Musiclink(TypedDict):
    musicLinks: Dict


class Shoplink(TypedDict):
    shopLinks: Dict


class Marginblock(TypedDict):
    marginBlocks: Dict


class Creatordetaileditlinks(TypedDict):
    isLoading: bool
    error: None
    editingProfileLinks: List
    isActiveUrlPastingOnText: bool
    profileLinkWidth: int
    profileLinkErrors: List
    multipleImageLinkIndex: int
    fourImageLinkIndex: int


class Fontcolor(TypedDict):
    r: int
    g: int
    b: int
    a: int


class Creatordetaileditprofile(TypedDict):
    isLoading: bool
    error: None
    editingCreatorDetailLayout: None
    selectedBackgroundCategory: str
    fontColor: Fontcolor
    backgroundColor: Fontcolor
    backgroundGradationStartColor: Fontcolor
    backgroundGradationEndColor: Fontcolor
    backgroundGradationColorPaletteIndex: int
    linkColor: Fontcolor
    backgroundImageUrlForOverlay: str


class Creatordetaileditgenre(TypedDict):
    error: None
    isLoading: bool
    selectedSnsActivityGenreId: None


class Snsactivitylist(TypedDict):
    error: None
    isLoading: bool


class Creatorpreference(TypedDict):
    creatorPreferences: Dict


class Notification(TypedDict):
    notifications: Dict


class Notificationrelationship(TypedDict):
    notificationRelationships: Dict
    totalNotificationUnreadCount: int


class Initialstate(TypedDict):
    account: Account
    creatorDetailEdit: Creatordetailedit
    genreCategory: Genrecategory
    profile: Profile0
    lineLogin: Linelogin
    login: Login
    modal: Modal
    lineMessaging: Linemessaging
    passwordReminder: Passwordreminder
    signUp: Signup
    firebaseAuth: Firebaseauth
    signupDetail: Signupdetail
    lineSignup: Linesignup
    analytics: Analytics
    notificationList: Notificationlist
    creatorDetailEditTutorial: Creatordetailedittutorial
    accountDelete: Accountdelete
    profileImageNFTModal: Profileimagenftmodal
    signupGenre: Signupgenre
    maintenance: Maintenance
    authentication: Authentication
    urlAlertModal: Urlalertmodal
    profileSnsModal: Profilesnsmodal
    creator: Creator
    creatorDetailLayout: Creatordetaillayout0
    creatorSnsActivityCategory: Creatorsnsactivitycategory
    snsIconLink: Snsiconlink
    user: User
    buttonLink: Buttonlink0
    textLink: Textlink1
    imageLink: Imagelink1
    movieLink: Movielink
    musicLink: Musiclink
    shopLink: Shoplink
    marginBlock: Marginblock
    creatorDetailEditLinks: Creatordetaileditlinks
    selectBackgroundImageModal: Profileimagenftmodal
    creatorDetailEditProfile: Creatordetaileditprofile
    creatorDetailEditGenre: Creatordetaileditgenre
    snsActivityList: Snsactivitylist
    creatorPreference: Creatorpreference
    editingMarginBlock: Profileimagenftmodal
    notification: Notification
    notificationRelationship: Notificationrelationship


class Pageprops(TypedDict):
    profile: Profile
    ogpImageUrl: str
    errorCode: str
    initialState: Initialstate


class Props(TypedDict):
    pageProps: Pageprops
    __N_SSP: bool


class Query(TypedDict):
    creatorUrl: str


class Root(TypedDict):
    props: Props
    page: str
    query: Query
    buildId: str
    isFallback: bool
    gssp: bool
    locale: str
    locales: List[str]
    defaultLocale: str
    scriptLoader: List



import pytest
from iwashi.visitor import Result
from iwashi.service.bandcamp import Bandcamp
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_bandcamp():
    service = Bandcamp()
    correct = Result(
        service=service,
        id="toxicholocaust",
        url="https://toxicholocaust.bandcamp.com",
        name="Toxic Holocaust",
        description="Toxic Thrash Metal\nEst. 1999",
        profile_picture="https://f4.bcbits.com/img/0032604396_21.jpg",
        links={"http://toxicholocaust.com"},
    )
    await _test_service(service, correct, "https://toxicholocaust.bandcamp.com")



from __future__ import annotations
import asyncio
import random
import re
import string
import time
from typing import Any, Callable

from loguru import logger

from .visitor import Result

USER_AGENT = "Profile Link Generator (https://github.com/am230/iwashi)"
BASE_HEADERS = {"User-Agent": USER_AGENT}
HTTP_REGEX = "(https?://)?(www.)?"
DEBUG = False


def print_result(
    result: Result, indent_level=0, print: Callable[[str], Any] = print
) -> None:
    indent = indent_level * "    "
    print(f"{indent}{result.service.name}")
    print(f"{indent}|id    : {result.id}")
    print(f"{indent}│url   : {result.url}")
    print(f"{indent}│name  : {result.name}")
    print(f"{indent}│links : {result.links}")
    if result.description:
        print(f"{indent}│description: " + result.description.replace("\n", "\\n"))
    for child in result.children:
        print_result(child, indent_level + 1, print)


def random_string(length: int) -> str:
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


URL_NORMALIZE_REGEX = r"(?P<protocol>https?)?:?\/?\/?(?P<domain>[^.]+\.[^\/]+)(?P<path>[^?#]+)?(?P<query>.+)?"


def normalize_url(url: str, https: bool = True) -> str | None:
    url = str(url).strip()
    match = re.match(URL_NORMALIZE_REGEX, url)
    if match is None:
        return None
    protocol = match.group("protocol") or "https"
    domain = match.group("domain")
    path = match.group("path") or ""
    query = match.group("query") or ""
    if https:
        protocol = "https"
    return f"{protocol}://{domain}{path}{query}"


def retry(
    max_retry: int,
    retry_interval: int = 0,
    retry_on: Callable[[Exception], bool] = lambda _: True,
) -> Callable[..., Any]:
    def decorator(func):
        def wrapper(*args, **kwargs):
            for i in range(max_retry):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if not retry_on(e):
                        raise e
                    logger.debug(f"Retrying {func.__name__} ({i + 1}/{max_retry})")
                    time.sleep(retry_interval)
                    continue
            raise Exception(
                f"Failed to execute {func.__name__} after {max_retry} retries"
            )

        return wrapper

    return decorator


def retry_async(
    max_retry: int,
    retry_interval: int = 0,
    retry_on: Callable[[Exception], bool] = lambda _: True,
) -> Callable[..., Any]:
    def decorator(func):
        async def wrapper(*args, **kwargs):
            e: Exception | None = None
            for i in range(max_retry):
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    if not retry_on(e):
                        raise e
                    logger.debug(f"Retrying {func.__name__} ({i + 1}/{max_retry})")
                    await asyncio.sleep(retry_interval)
                    continue
            raise Exception(
                f"Failed to execute {func.__name__} after {max_retry} retries"
            )

        return wrapper

    return decorator


def cache(func: Callable[..., Any]) -> Callable[..., Any]:
    cache = {}

    def wrapper(*args, **kwargs):
        key = (args, tuple(kwargs.items()))
        if key in cache:
            return cache[key]
        result = func(*args, **kwargs)
        cache[key] = result
        return result

    return wrapper


def cache_async(func: Callable[..., Any]) -> Callable[..., Any]:
    cache = {}

    async def wrapper(*args, **kwargs):
        key = (args, tuple(kwargs.items()))
        if key in cache:
            return cache[key]
        result = await func(*args, **kwargs)
        cache[key] = result
        return result

    return wrapper


class Traverser[T]:
    def __init__(self, value: T | None = None):
        self.value = value

    def map[V](self, func: Callable[[T], V], default: V | None = None) -> Traverser[V]:
        if self.value is None:
            return TRAVERSE_NONE
        value = func(self.value)
        if value is None:
            return Traverser(default)
        return Traverser(value)

    def get(self, default: T | None = None) -> T | None:
        if self.value is None:
            return default
        return self.value

    def unwrap(self) -> T:
        if self.value is None:
            raise ValueError("Value is None")
        return self.value

    def is_none(self) -> bool:
        return self.value is None


TRAVERSE_NONE = Traverser(None)


def traverse[T](value: T | None) -> Traverser[T]:
    """
    Traverse a value that may be None.
    """
    return Traverser(value)



from __future__ import annotations
from typing import Dict, List, NotRequired, TypedDict

import json
import re

import bs4

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class Sketch(Service):
    def __init__(self):
        super().__init__(
            name="Sketch",
            regex=re.compile(
                HTTP_REGEX + r"sketch\.pixiv\.net/@(?P<id>\w+)", re.IGNORECASE
            ),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://sketch.pixiv.net/@{id}"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        element = soup.select_one("script#__NEXT_DATA__")
        if element is None:
            raise Exception(f"__NEXT_DATA__ not found: {id}")
        next_data = json.loads(element.text)
        data: Root = json.loads(next_data["props"]["pageProps"]["initialState"])
        users = data["users"]["users"]
        if len(users) != 1:
            raise Exception(f"User is must be unique: {id}")
        user = users.popitem()[1]
        context.create_result(
            self,
            id=id,
            url=url,
            name=user["name"],
            description=user["description"],
            profile_picture=user["icon"]["photo"]["original"]["url"],
        )
        social_accounts = user["social_accounts"]
        if "twitter" in social_accounts:
            unique_name = social_accounts["twitter"]["unique_name"]
            context.enqueue_visit(f"https://twitter.com/{unique_name}")
            del social_accounts["twitter"]
        if "pixiv" in social_accounts:
            unique_name = social_accounts["pixiv"]["unique_name"]
            context.enqueue_visit(f"https://www.pixiv.net/users/{unique_name}")
            del social_accounts["pixiv"]
        if social_accounts:
            raise Exception(f"Unknown social accounts: {social_accounts}")


class Amountpoint(TypedDict):
    partial: bool
    input: str
    exchangeErrors: List


class Amountmoney(TypedDict):
    partial: bool
    input: str
    transferErrors: List


class Transfer(TypedDict):
    to: str
    amountPoint: Amountpoint
    amountMoney: Amountmoney
    confirmation: None


class Reward(TypedDict):
    balance: Dict
    provisionalRewards: Dict
    decrementHistories: Dict
    rewards: Dict
    rewardWallLink: Dict
    bankAccount: Dict
    transferring: bool
    historying: bool
    transfer: Transfer
    terms: None


class Temporarymedium(TypedDict):
    temporaryMedium: Dict


class Template(TypedDict):
    themes: Dict


class State(TypedDict):
    state: Dict


class Special(TypedDict):
    contestResults: Dict


class Feedback(TypedDict):
    feedbacks: Dict
    links: Dict


class Contest(TypedDict):
    contests: List


class Info(TypedDict):
    content: None
    hidden: bool


class Componentactionerror(TypedDict):
    alert: None
    info: Info


class Options(TypedDict):
    filterBy: None
    orderBy: str


class Livelist(TypedDict):
    ids: List
    ended: bool
    fetching: bool
    options: Options


class Live(TypedDict):
    lives: Dict
    liveAdStages: Dict
    liveAvailability: bool
    liveClosedAvailability: bool
    liveLogs: Dict
    liveCaptions: Dict
    liveChatErrorMessage: None
    liveArchiveLink: Dict
    liveGiftings: Dict
    liveGiftingSummaries: Dict
    liveGiftingRecommends: Dict
    liveGiftingRecommendsPaging: Dict
    liveGiftingHistories: Dict
    liveSummaryUrl: None
    liveList: Livelist
    unpresentedBlacklistedLiveChats: List


class Comment(TypedDict):
    comments: Dict
    commentWalls: Dict
    commentLinks: Dict
    heartWalls: Dict
    heartLinks: Dict


class Tag(TypedDict):
    endedTagCompletions: List
    tags: List
    tagWalls: Dict
    tagLinks: Dict
    tagUploadSuggested: Dict


class Item(TypedDict):
    items: Dict
    itemWalls: Dict
    itemLinks: Dict
    itemRepliesWall: Dict
    itemRepliesLink: Dict
    itemTopicIds: List


class Browser(TypedDict):
    name: str
    version: str
    major: str
    native: bool
    uwp: bool


class Engine(TypedDict):
    name: str
    version: str


class Cpu(TypedDict):
    architecture: str


class Agent(TypedDict):
    ua: str
    browser: Browser
    engine: Engine
    os: Engine
    device: Dict
    cpu: Cpu


class App(TypedDict):
    agent: Agent
    locale: str
    mayPromptNotificationSubscription: bool
    renderingPreparedAt: int
    preparedSimpleDrawOnClient: bool
    uploadables: List[str]
    isInit: bool


class Gifting(TypedDict):
    lastGiftingItem: None
    giftingItems: List
    points: List


class Contact(TypedDict):
    faqCategories: List


class Announcement(TypedDict):
    announcements: List


class DescriptionFragmentsItem(TypedDict):
    type: str
    body: str
    normalized_body: str


class Twitter(TypedDict):
    unique_name: str
    expired: bool
    is_public: bool


class Pixiv(TypedDict):
    unique_name: str
    expired: bool
    show_on_pixiv: bool


class SocialAccounts(TypedDict):
    twitter: NotRequired[Twitter]
    pixiv: NotRequired[Pixiv]


class Color(TypedDict):
    hex: str
    r: int
    g: int
    b: int


class Pxsq60(TypedDict):
    width: int
    height: int
    url: str
    url2x: str


class Photo(TypedDict):
    pxsq60: Pxsq60
    pxw540: Pxsq60
    pxsq180: Pxsq60
    pxsq120: Pxsq60
    sq180: Pxsq60
    original: Pxsq60
    sq120: Pxsq60
    w240: Pxsq60
    sq60: Pxsq60
    w540: Pxsq60
    pxw240: Pxsq60


class Icon(TypedDict):
    id: int
    type: str
    color: Color
    photo: Photo


class Self(TypedDict):
    method: str
    href: str


class _Links(TypedDict):
    self: Self


class Stats(TypedDict):
    follower_count: int
    following_count: int
    heart_count: int
    resnap_count: int
    public_post_count: int


class User(TypedDict):
    description_fragments: List[DescriptionFragmentsItem]
    name: str
    followed: bool
    following: bool
    blocking: bool
    social_accounts: SocialAccounts
    icon: Icon
    unique_name: str
    post_ids: List
    _links: _Links
    id: str
    pixiv_user_id: str
    description: str
    stats: Stats


Users = Dict[str, User]
Usercolors = Dict[str, str]


class Users0(TypedDict):
    users: Users
    userWalls: Dict
    userLinks: Dict
    userColors: Usercolors
    usersRecommended: List
    usersBlockedWall: List
    usersBlockedWallLink: Dict
    account: Dict
    accountSetting: Dict


class Notifications(TypedDict):
    wall: List
    link: Dict
    types: List


class Privacypolicy(TypedDict):
    privacyPolicyIsAccepted: None
    privacyPolicyText: str
    privacyPolicyUrl: str
    privacyPolicyAcceptedVersion: str
    privacyPolicyUpdatedAt: str


class Params(TypedDict):
    id: str


class Navigation(TypedDict):
    name: str
    params: Params
    search: Dict


class Status(TypedDict):
    code: str
    message: None


class Route(TypedDict):
    historyState: None
    isNavigating: bool
    navigation: Navigation
    status: Status


class Root(TypedDict):
    reward: Reward
    temporaryMedium: Temporarymedium
    template: Template
    state: State
    special: Special
    feedback: Feedback
    contest: Contest
    componentActionError: Componentactionerror
    live: Live
    comment: Comment
    tag: Tag
    item: Item
    app: App
    gifting: Gifting
    contact: Contact
    announcement: Announcement
    users: Users0
    notifications: Notifications
    privacyPolicy: Privacypolicy
    route: Route


class Pageprops(TypedDict):
    initialState: str


class Props(TypedDict):
    pageProps: Pageprops
    __N_SSP: bool


class Query(TypedDict):
    id: str


class NEXT_DATA(TypedDict):
    props: Props
    page: str
    query: Query
    buildId: str
    isFallback: bool
    gssp: bool
    customServer: bool
    scriptLoader: List



import pytest
from iwashi.service.nicovideo import Nicovideo
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_nicovideo():
    service = Nicovideo()
    correct = Result(
        service=service,
        id="128134532",
        url="https://www.nicovideo.jp/user/128134532",
        name="ラベンダーP",
        description="岡山県備前市日生町に住む2006年生まれの人 <br>好きなもの（鉄道、東方、艦これ、ボカロ、ボイロ、野球、中日、阪神、西武、オリックス、ロッテ、railwars、すくすく白沢）",
        profile_picture="https://secure-dcdn.cdn.nimg.jp/nicoaccount/usericon/12813/128134532.jpg?1710066983",
        links={
            "https://www.youtube.com/channel/UC4jehnRY1GBPBpg-Np4WFNg",
            "https://twitter.com/lavenderp2018",
        },
    )
    await _test_service(service, correct, "https://www.nicovideo.jp/user/128134532")



from typing import Iterable

import aiohttp

from iwashi.helper import BASE_HEADERS
from iwashi.visitor import Context, FakeVisitor, Result, Service


def iterable_eq(a: Iterable, b: Iterable) -> bool:
    # Compare two iterables for equality
    # ignoring the order of elements
    for item in a:
        if item not in b:
            return False
    for item in b:
        if item not in a:
            return False
    return True


async def _test_service(service: Service, correct_result: Result, *urls: str) -> None:
    # resolve id
    visitor = FakeVisitor()
    session = aiohttp.ClientSession(headers=BASE_HEADERS)
    for url in urls:
        context = Context(session=session, visitor=visitor)
        resolved_id = await service.resolve_id(context, url)
        assert (
            resolved_id == correct_result.id
        ), f"ID mismatch for {url} ({resolved_id} != {correct_result.id})"

    # visit
    for url in urls:
        result = await service.visit_url(session, url)
        assert result, f"Failed to visit {url}"
        assert (
            result.url == correct_result.url
        ), f"URL mismatch for {url} ({result.url} != {correct_result.url})"
        assert (
            result.service == correct_result.service
        ), f"Service mismatch for {url} ({result.service} != {correct_result.service})"
        assert (
            result.name == correct_result.name
        ), f"Name mismatch for {url} ({result.name} != {correct_result.name})"
        assert (
            result.description == correct_result.description
        ), f"Description mismatch for {url} ({result.description} != {correct_result.description})"
        assert (
            correct_result.profile_picture
            and result.profile_picture
            and (correct_result.profile_picture in result.profile_picture)
        ), f"Profile picture mismatch for {url} ({result.profile_picture} != {correct_result.profile_picture})"
        assert iterable_eq(
            result.links, correct_result.links
        ), f"Links mismatch for {url} ({result.links} != {correct_result.links})"



import re

import bs4
from loguru import logger

from iwashi.helper import HTTP_REGEX, normalize_url
from iwashi.visitor import Context, Service


class TwitCasting(Service):
    def __init__(self) -> None:
        super().__init__(
            name="TwitCasting",
            regex=re.compile(
                HTTP_REGEX + r"twitcasting\.tv/(?P<id>[-\w]+)", re.IGNORECASE
            ),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://twitcasting.tv/{id}"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        element = soup.select_one(".tw-user-nav-name")
        if element is None:
            logger.warning(f"[TwitCasting] Could not find name for {url}")
            return
        name = element.text.strip()
        element = soup.select_one(".tw-user-nav-icon > img")
        profile_picture = None
        if element is not None:
            attr = element["src"]
            if isinstance(attr, str):
                profile_picture = normalize_url(attr)
        description: str | None = None
        element = soup.select_one('meta[name="description"]')
        if element is not None:
            description = element.attrs.get("content")

        context.create_result(
            self,
            id=id,
            url=url,
            name=name,
            description=description,
            profile_picture=profile_picture,
        )

        links = set()  # TODO
        for element in soup.select(".tw-follow-list-row-icon"):
            links.add(element["href"])
        for link in links:
            if link.startswith("/"):
                continue
            context.enqueue_visit(link)



import asyncio

import click

from .helper import print_result
from .iwashi import tree


@click.command()
@click.argument("url", required=True)
def main(url: str) -> None:
    result = asyncio.run(tree(url))
    assert result
    print("\n" * 4)
    print_result(result)


if __name__ == "__main__":
    main()



import re
from typing import List, TypedDict

from loguru import logger

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class Instagram(Service):
    def __init__(self) -> None:
        super().__init__(
            name="Instagram",
            regex=re.compile(HTTP_REGEX + r"instagram\.com/(?P<id>\w+)", re.IGNORECASE),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://www.instagram.com/{id}"
        res = await context.session.get(url)
        res.raise_for_status()
        match = re.search(r"\"X-IG-App-ID\": ?\"(?P<id>\d{15})\"", await res.text())
        if match is None:
            raise Exception(f"[Instagram] No X-IG-App-ID found in {url}")
        context.session.headers["x-ig-app-id"] = match.group("id")

        info_res = await context.session.get(
            f"https://www.instagram.com/api/v1/users/web_profile_info/?username={id}",
        )
        info_res.raise_for_status()
        if info_res.status // 100 != 2 or info_res.history:
            logger.warning("[Instagram] Blocked by Instagram")
            context.create_result(
                self,
                id=id,
                url=url,
                name=id,
                description="Blocked by Instagram",
            )
            return
        info: Root = await info_res.json()
        user = info["data"]["user"]
        context.create_result(
            self,
            id=id,
            url=url,
            name=user["full_name"],
            profile_picture=user["profile_pic_url"],
            description=user["biography"],
        )

        for link in user["bio_links"]:
            context.enqueue_visit(link["url"])


class FriendshipStatus(TypedDict):
    following: bool
    blocking: bool
    is_bestie: bool
    is_feed_favorite: bool
    is_restricted: bool
    muting: bool
    is_muting_reel: bool
    outgoing_request: bool
    followed_by: bool
    incoming_request: bool


class HdProfilePicUrlInfo(TypedDict):
    url: str


class BiographyWithEntities(TypedDict):
    entities: List


class BioLinksItem(TypedDict):
    link_type: str
    lynx_url: str
    title: str
    url: str


class User(TypedDict):
    friendship_status: FriendshipStatus
    full_name: str
    gating: None
    is_memorialized: bool
    is_private: bool
    has_story_archive: None
    username: str
    is_regulated_c18: bool
    regulated_news_in_locations: List
    text_post_app_badge_label: str
    show_text_post_app_badge: bool
    eligible_for_text_app_activation_badge: bool
    hide_text_app_activation_badge_on_text_app: bool
    pk: str
    live_broadcast_visibility: None
    live_broadcast_id: None
    profile_pic_url: str
    hd_profile_pic_url_info: HdProfilePicUrlInfo
    is_unpublished: bool
    mutual_followers_count: int
    profile_context_links_with_user_ids: List
    biography_with_entities: BiographyWithEntities
    account_badges: List
    bio_links: List[BioLinksItem]
    external_lynx_url: str
    external_url: str
    ai_agent_type: None
    has_chaining: bool
    fbid_v2: str
    supervision_info: None
    interop_messaging_user_fbid: str
    account_type: int
    biography: str
    is_embeds_disabled: bool
    show_account_transparency_details: bool
    is_verified: bool
    is_professional_account: None
    follower_count: int
    address_street: str
    city_name: str
    is_business: bool
    zip: str
    category: str
    should_show_category: bool
    pronouns: List
    transparency_label: None
    transparency_product: None
    following_count: int
    media_count: int
    latest_reel_media: int
    total_clips_count: int
    latest_besties_reel_media: int
    reel_media_seen_timestamp: int
    id: str


class User0(TypedDict):
    pk: str
    id: str
    can_see_organic_insights: bool
    has_onboarded_to_text_post_app: bool


class Viewer(TypedDict):
    user: User0


class Data(TypedDict):
    user: User
    viewer: Viewer


class Extensions(TypedDict):
    is_final: bool


class Root(TypedDict):
    data: Data
    extensions: Extensions



import pytest
from iwashi.service.sketch import Sketch
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_sketch():
    service = Sketch()
    correct = Result(
        service=service,
        id="par1y",
        url="https://sketch.pixiv.net/@par1y",
        name="Paryi",
        description="お勉強と落書きがメイン\nフォロワーさんとまったり絡みたい\nhttps://twitter.com/par1y\n",
        profile_picture="https://img-sketch.pixiv.net/uploads/user_icon/file/1980676/5116648097160323811.jpg",
        links={"https://twitter.com/par1y", "https://www.pixiv.net/users/par1y"},
    )
    await _test_service(service, correct, "https://sketch.pixiv.net/@par1y")



from __future__ import annotations
import json

import re
from typing import Dict, List, Literal, TypedDict

import bs4

from iwashi.helper import HTTP_REGEX, BASE_HEADERS
from iwashi.visitor import Context, Service


class Spotify(Service):
    def __init__(self):
        super().__init__(
            name="Spotify",
            regex=re.compile(
                HTTP_REGEX
                + r"(open\.)?spotify\.com\/(intl-[\w]*\/)?artist\/(?P<id>[0-9a-zA-Z]+)",
                re.IGNORECASE,
            ),
        )

    async def extract_access_token(self, soup: bs4.BeautifulSoup) -> str:
        script = soup.select_one("script#session")
        if script is None:
            raise RuntimeError("Token not found")
        data = Session(**json.loads(script.text))
        return data["accessToken"]

    async def visit(self, context: Context, id: str):
        url = f"https://open.spotify.com/artist/{id}"
        response = await context.session.get(url, headers=BASE_HEADERS)
        response.raise_for_status()
        soup = bs4.BeautifulSoup(await response.text(), "html.parser")
        access_token = await self.extract_access_token(soup)

        headers = {
            "accept": "application/json",
            "authorization": f"Bearer {access_token}",
            "content-type": "application/json",
        }
        params = {
            "operationName": "queryArtistOverview",
            "variables": json.dumps(
                {
                    "uri": f"spotify:artist:{id}",
                    "locale": "intl-ja",
                    "includePrerelease": True,
                }
            ),
            "extensions": '{"persistedQuery":{"version":1,"sha256Hash":"da986392124383827dc03cbb3d66c1de81225244b6e20f8d78f9f802cc43df6e"}}',
        }
        response = await context.session.get(
            "https://api-partner.spotify.com/pathfinder/v1/query",
            params=params,
            headers=headers | BASE_HEADERS,
        )
        response.raise_for_status()
        data: ArtistResponse = await response.json()
        profile = data["data"]["artistUnion"]["profile"]
        avatar_image = data["data"]["artistUnion"]["visuals"]["avatarImage"]
        if len(avatar_image["sources"]) == 0:
            avatar_url = None
        else:
            avatar_url = sorted(
                avatar_image["sources"], key=lambda x: x["width"] * x["height"]
            )[-1]["url"]

        context.create_result(
            self,
            id=id,
            url=url,
            name=profile["name"],
            description=profile["biography"]["text"],
            profile_picture=avatar_url,
        )

        for link in profile["externalLinks"]["items"]:
            context.enqueue_visit(link["url"])


class Session(TypedDict):
    accessToken: str  # [0-9a-zA-Z-_]{19}-[0-9a-zA-Z-_]{37}-[0-9a-zA-Z-_]{57}
    accessTokenExpirationTimestampMs: int  # Unix timestamp in milliseconds
    isAnonymous: bool  # true
    clientId: str  # [0-9a-f]{32}


class Biography(TypedDict):
    text: str
    title: Literal["AUTOBIOGRAPHY"]


class ExternalLinksItem(TypedDict):
    name: str
    url: str


class ExternalLinks(TypedDict):
    items: List[ExternalLinksItem]


class ImagesItemSource(TypedDict):
    width: int
    height: int
    url: str


class ImagesItem(TypedDict):
    sources: List[ImagesItemSource]


class Images(TypedDict):
    items: List[ImagesItem]


class OwnerV2Data(TypedDict):
    __typename: Literal["User"]
    name: str


class OwnerV2(TypedDict):
    data: OwnerV2Data


class PlaylistsV2ItemData(TypedDict):
    __typename: Literal["Playlist"]
    description: str
    images: Images
    name: str
    ownerV2: OwnerV2
    uri: str


class PlaylistsV2Item(TypedDict):
    data: PlaylistsV2ItemData


class PlaylistsV2(TypedDict):
    items: List[PlaylistsV2Item]


class Profile(TypedDict):
    biography: Biography
    externalLinks: ExternalLinks
    name: str
    pinnedItem: None
    playlistsV2: PlaylistsV2
    verified: bool


class ColorRaw(TypedDict):
    hex: str  # "#[0-9a-f]{6}"


class ExtractedColors(TypedDict):
    colorRaw: ColorRaw


class AvatarImage(ImagesItem):
    extractedColors: ExtractedColors


class Visuals(TypedDict):
    avatarImage: AvatarImage


class ArtistUnion(TypedDict):
    profile: Profile
    visuals: Visuals


class ArtistData(TypedDict):
    artistUnion: ArtistUnion


class ArtistResponse(TypedDict):
    data: ArtistData
    extensions: Dict



import pytest
from iwashi.service.soundcloud import Soundcloud
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_soundcloud():
    service = Soundcloud()
    correct = Result(
        service=service,
        id="speder2",
        url="https://soundcloud.com/speder2",
        name="Speder2",
        description="ゲーム製作サークルの音楽・SEを担当しています",
        profile_picture="https://i1.sndcdn.com/avatars-000125467778-jpowym-large.jpg",
        links={
            "http://kohada.ushimairi.com/",
        },
    )
    await _test_service(
        service,
        correct,
        "https://soundcloud.com/speder2",
        "https://soundcloud.com/speder2/tracks",
    )



from __future__ import annotations

import re
from typing import Any, List, Literal, NotRequired, TypedDict

from loguru import logger


from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class Skeb(Service):
    def __init__(self):
        super().__init__(
            name="Skeb",
            regex=re.compile(HTTP_REGEX + r"skeb\.jp/@(?P<id>[\w-]+)", re.IGNORECASE),
        )
        self.request_key: str | None = None

    async def fetch_request_key(self, context: Context, url: str) -> str:
        if self.request_key is not None:
            return self.request_key
        res = await context.session.get(url)
        request_key = res.cookies.get("request_key")
        if request_key is None:
            raise RuntimeError("Could not find request_key")
        self.request_key = request_key.value
        return self.request_key

    async def visit(self, context: Context, id: str):
        url = f"https://skeb.jp/@{id}"
        request_key = await self.fetch_request_key(context, url)
        api_endpoint = f"https://skeb.jp/api/users/{id}"

        cookies = {
            "request_key": request_key,
        }

        headers = {
            "accept": "application/json, text/plain, */*",
            "accept-language": "ja",
            "authorization": "Bearer null",
            "cache-control": "no-cache",
            "pragma": "no-cache",
            "priority": "u=1, i",
            "referer": url,
        }
        res = await context.session.get(api_endpoint, cookies=cookies, headers=headers)
        res.raise_for_status()
        data: Root = await res.json()

        context.create_result(
            self,
            id=id,
            url=url,
            name=data["name"],
            description=data["description"],
            profile_picture=data["avatar_url"],
        )

        context.enqueue_visit(data["url"])

        NOT_SUPPORTED = None
        PROVIDERS = {
            "nijie_id": NOT_SUPPORTED,
            "dlsite_id": NOT_SUPPORTED,
            "fanza_id": NOT_SUPPORTED,
            "pixiv_id": "https://www.pixiv.net/member.php?id={id}",
            "booth_id": "https://{id}.booth.pm",
            "fantia_id": "https://fantia.jp/fanclubs/{id}",
            "fanbox_id": "https://{id}.fanbox.cc",
            "skima_id": "https://skima.jp/profile?id={id}",
            "coconala_id": "https://coconala.com/users/{id}",
            "patreon_id": "https://www.patreon.com/{id}",
            "wishlist_id": NOT_SUPPORTED,
            "youtube_id": "https://www.youtube.com/channel/{id}",
        }

        for key, template in PROVIDERS.items():
            if key not in data:
                logger.warning(f"Could not find {key} in {data}")
                continue
            value = data[key]
            if value is None:
                continue
            if template is NOT_SUPPORTED:
                logger.warning(f"{key} is not supported")
                continue
            formatted_url = template.format(id=value)
            context.enqueue_visit(formatted_url)

        SERVICES = {
            "twitter": "https://twitter.com/{screen_name}",
        }
        for user_service_link in data.get("user_service_links", {}):
            if user_service_link["provider"] not in SERVICES:
                logger.warning(
                    f"Could not find {user_service_link['provider']} in {data}"
                )
                continue
            formatted_url = SERVICES[user_service_link["provider"]].format(
                screen_name=user_service_link["screen_name"]
            )
            context.enqueue_visit(formatted_url)


type UnknownField = Any
type UnknownId = int | str


class UserServiceLinksItem(TypedDict):
    provider: str
    screen_name: str


class SkillsItem(TypedDict):
    genre: str
    default_amount: NotRequired[int]


class CreditCardsItem(TypedDict):
    nsfw: bool


class CreditCards(TypedDict):
    jcb: CreditCardsItem
    master_card: CreditCardsItem
    visa: CreditCardsItem
    discover: CreditCardsItem
    american_express: CreditCardsItem


class PaymentMethods(TypedDict):
    credit_cards: CreditCards


class ImageUrls(TypedDict):
    src: str
    srcset: str


type Genre = Literal["art"]


class WorksItem(TypedDict):
    path: str
    private_thumbnail_image_urls: ImageUrls
    private: bool
    genre: Genre
    tipped: bool
    creator_id: int
    client_id: int
    vtt_url: UnknownField
    thumbnail_image_urls: ImageUrls
    duration: UnknownField
    nsfw: bool
    hardcore: bool
    consored_thumbnail_image_urls: ImageUrls
    body: str
    word_count: int
    transcoder: Literal["image"]
    creator_acceptable_same_genre: bool


class SimilarCreatorsItem(TypedDict):
    id: int
    creator: bool
    nsfw_acceptable: bool
    acceptable: bool
    name: str
    screen_name: str
    avatar_url: str
    header_url: str
    appeal_receivable: bool
    popular_creator_rank: UnknownField
    request_master_rank: UnknownField
    first_requester_rank: UnknownField
    deleted_at: UnknownField
    tip_acceptable_by: Literal["all"]
    accept_expiration_days: int
    skills: List[SkillsItem]
    genre: Genre


class Root(TypedDict):
    id: int
    acceptable: bool
    creator: bool
    avatar_url: str
    name: str
    nsfw_acceptable: bool
    private_acceptable: bool
    screen_name: str
    language: str
    header_url: str
    body_size: str
    received_works_count: int
    received_private_works_count: int
    received_nsfw_works_count: int
    sent_requests_average_cancel_time: int
    appeal_receivable: bool
    request_master_rank: UnknownField
    first_requester_rank: UnknownField
    sent_first_works_count: int
    sent_public_works_count: int
    popular_creator_rank: UnknownField
    instruction: Literal["unspecified"]
    asct: UnknownField
    description: str
    nijie_id: UnknownId
    dlsite_id: UnknownId
    fanza_id: UnknownId
    pixiv_id: int
    booth_id: UnknownId
    fantia_id: UnknownId
    fanbox_id: UnknownId
    skima_id: UnknownId
    coconala_id: UnknownId
    patreon_id: UnknownId
    busy: bool
    url: str
    wishlist_id: UnknownField
    youtube_id: str
    complete_rate: float
    show_social_profile: bool
    accept_expiration_days: int
    complete_expiration_days: int
    twitter_uid: str
    user_service_links: List[UserServiceLinksItem]
    banned: bool
    banned_by: str
    og_image_url: str
    skills: List[SkillsItem]
    payment_methods: PaymentMethods
    genre: Genre
    default_amount: UnknownField
    acceptable_only_private: bool
    tip_acceptable: bool
    yearly_report_client_visible: bool
    received_works: List[WorksItem]
    sent_works: List[WorksItem]
    similar_creators: List[SimilarCreatorsItem]



import pytest
from iwashi.service.skeb import Skeb
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_skeb():
    service = Skeb()
    correct = Result(
        service=service,
        id="adultfear",
        url="https://skeb.jp/@adultfear",
        name="nerrin ☀️",
        description="Nerr (she/they) 20+ 🌺 EN/中文/日本語・Writer, GM, sometimes Artist・OCs, d&d, ttrpg, type-moon, toku, horror・yellow 🧡💛",
        profile_picture="https://pbs.twimg.com/profile_images/1648903417262403584/IOswj0Xs.jpg",
        links={
            "https://twitter.com/adultfear",
            "http://adultfear.carrd.co",
        },
    )
    await _test_service(
        service,
        correct,
        "https://skeb.jp/@adultfear",
    )



import re
from typing import List, TypedDict

from loguru import logger

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class Twitch(Service):
    def __init__(self) -> None:
        super().__init__(
            name="Twitch",
            regex=re.compile(HTTP_REGEX + r"twitch\.tv/(?P<id>\w+)", re.IGNORECASE),
        )

    async def visit(self, context: Context, id: str):
        url = f"https://www.twitch.tv/{id}"
        res = await context.session.get(url)
        res.raise_for_status()
        match = re.search(
            r'clientId ?= ?"(?P<token>kimne78kx3ncx6brgo4mv6wki5h1ko)"',
            await res.text(),
        )
        if match is None:
            logger.warning(f"[Twitch] Could not find token for {url}")
            return
        token = match.group("token")

        res = await context.session.post(
            "https://gql.twitch.tv/gql",
            json=[
                {
                    "operationName": "ChannelRoot_AboutPanel",
                    "variables": {"channelLogin": id, "skipSchedule": True},
                    "extensions": {
                        "persistedQuery": {
                            "version": 1,
                            "sha256Hash": "6089531acef6c09ece01b440c41978f4c8dc60cb4fa0124c9a9d3f896709b6c6",
                        }
                    },
                }
            ],
            headers={
                "Client-Id": token,
            },
        )
        res.raise_for_status()
        info: Root = await res.json()
        user = info[0]["data"]["user"]
        if user is None:
            logger.warning(f"[Twitch] Could not find user for {url}")
            return
        context.create_result(
            self,
            id=id,
            url=url,
            name=user["displayName"],
            description=user["description"],
            profile_picture=user["profileImageURL"],
        )

        for link in user["channel"]["socialMedias"]:
            context.enqueue_visit(link["url"])


class Followers(TypedDict):
    totalCount: int
    __typename: str


class SocialMediasItem0(TypedDict):
    id: str
    name: str
    title: str
    url: str
    __typename: str


class Channel(TypedDict):
    id: str
    socialMedias: List[SocialMediasItem0]
    __typename: str


class LastBroadcast(TypedDict):
    id: str
    game: None
    __typename: str


class Videos(TypedDict):
    edges: List
    __typename: str


class User(TypedDict):
    id: str
    description: str
    displayName: str
    isPartner: bool
    primaryColorHex: str
    profileImageURL: str
    followers: Followers
    channel: Channel
    lastBroadcast: LastBroadcast
    primaryTeam: None
    videos: Videos
    __typename: str


class Data(TypedDict):
    currentUser: None
    user: User


class Extensions(TypedDict):
    durationMilliseconds: int
    operationName: str
    requestID: str


class RootItem0(TypedDict):
    data: Data
    extensions: Extensions


Root = List[RootItem0]



import json
import re
from typing import List, TypedDict

from loguru import logger

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service


class Twitter(Service):
    def __init__(self) -> None:
        super().__init__(
            name="Twitter",
            regex=re.compile(
                HTTP_REGEX + r"(twitter|x)\.com/(#!/)?@?(?P<id>\w+)", re.IGNORECASE
            ),
        )
        self.headers = {
            "accept": "*/*",
            "accept-encoding": "gzip, deflate, br",
            "accept-language": "ja",
        }
        self.bearer_token: str | None = None
        self.guest_token: str | None = None

    async def fetch_authorization(self, context: Context) -> str:
        if self.bearer_token:
            return self.bearer_token
        res = await context.session.get(
            "https://abs.twimg.com/responsive-web/client-web/main.28fc48ca.js"
        )
        res.raise_for_status()
        match = re.search('(AAAAA.*?)"', await res.text())
        assert match
        bearer_token = "Bearer " + match.group(1)
        self.bearer_token = bearer_token
        return bearer_token

    async def fetch_guest_token(self, context: Context) -> str:
        if self.guest_token:
            return self.guest_token
        res = await context.session.post(
            "https://api.twitter.com/1.1/guest/activate.json",
            headers=self.headers,
        )
        res.raise_for_status()
        data = await res.json()
        self.guest_token = data["guest_token"]
        return data["guest_token"]

    async def setup_headers(self, context: Context) -> None:
        self.headers["authorization"] = await self.fetch_authorization(context)
        self.headers["x-guest-token"] = await self.fetch_guest_token(context)

    async def visit(self, context: Context, id: str) -> None:
        url = f"https://twitter.com/{id}"
        await self.setup_headers(context)

        res = await context.session.get(
            "https://api.twitter.com/graphql/rePnxwe9LZ51nQ7Sn_xN_A/UserByScreenName",
            params={
                "variables": json.dumps(
                    {
                        "screen_name": id,
                        "withSafetyModeUserFields": True,
                        "withSuperFollowsUserFields": True,
                    }
                ),
                "features": json.dumps(
                    {
                        "responsive_web_twitter_blue_verified_badge_is_enabled": True,
                        "responsive_web_graphql_exclude_directive_enabled": True,
                        "verified_phone_label_enabled": False,
                        "responsive_web_graphql_skip_user_profile_image_extensions_enabled": False,
                        "responsive_web_graphql_timeline_navigation_enabled": True,
                    }
                ),
            },
            headers=self.headers,
        )
        res.raise_for_status()
        info: Root = await res.json()
        if not info["data"]:
            logger.warning(f"[Twitter] Could not find data for {url}")
            return
        result = info["data"]["user"]["result"]
        if result["__typename"] == "UserUnavailable":
            context.create_result(
                self,
                id=id,
                url=url,
                name="<UserUnavailable>",
            )
            return

        data = result["legacy"]
        context.create_result(
            self,
            id=id,
            url=url,
            name=data["name"],
            description=data["description"],
            profile_picture=data["profile_image_url_https"],
        )

        if "url" not in data["entities"]:
            return

        for link in data["entities"]["url"]["urls"]:
            context.enqueue_visit(link["expanded_url"])

        for link in data["entities"]["description"]["urls"]:
            context.enqueue_visit(link["expanded_url"])


class LocationsItem0(TypedDict):
    line: int
    column: int


class Tracing(TypedDict):
    trace_id: str


class Extensions(TypedDict):
    name: str
    source: str
    code: int
    kind: str
    tracing: Tracing


class ErrorsItem0(TypedDict):
    message: str
    locations: List[LocationsItem0]
    path: List[str]
    extensions: Extensions
    code: int
    kind: str
    name: str
    source: str
    tracing: Tracing


class AffiliatesHighlightedLabel(TypedDict):
    pass


class UrlsItem0(TypedDict):
    display_url: str
    expanded_url: str
    url: str
    indices: List[int]


class Description(TypedDict):
    urls: List[UrlsItem0]


class Entities(TypedDict):
    description: Description
    url: Description


class Rgb(TypedDict):
    blue: int
    green: int
    red: int


class PaletteItem0(TypedDict):
    percentage: float
    rgb: Rgb


class Ok(TypedDict):
    palette: List[PaletteItem0]


class R(TypedDict):
    ok: Ok


class MediaColor(TypedDict):
    r: R


class ProfileBannerExtensions(TypedDict):
    mediaColor: MediaColor


class Legacy(TypedDict):
    created_at: str
    default_profile: bool
    default_profile_image: bool
    description: str
    entities: Entities
    fast_followers_count: int
    favourites_count: int
    followers_count: int
    friends_count: int
    has_custom_timelines: bool
    is_translator: bool
    listed_count: int
    location: str
    media_count: int
    name: str
    normal_followers_count: int
    pinned_tweet_ids_str: List[str]
    possibly_sensitive: bool
    profile_banner_extensions: ProfileBannerExtensions
    profile_banner_url: str
    profile_image_extensions: ProfileBannerExtensions
    profile_image_url_https: str
    profile_interstitial_type: str
    screen_name: str
    statuses_count: int
    translator_type: str
    url: str
    verified: bool
    withheld_in_countries: List


class Result(TypedDict):
    __typename: str
    id: str
    rest_id: str
    affiliates_highlighted_label: AffiliatesHighlightedLabel
    is_blue_verified: bool
    legacy: Legacy
    business_account: AffiliatesHighlightedLabel
    legacy_extended_profile: AffiliatesHighlightedLabel
    is_profile_translatable: bool
    verification_info: AffiliatesHighlightedLabel


class User(TypedDict):
    result: Result


class Data(TypedDict):
    user: User


class Root(TypedDict):
    errors: List[ErrorsItem0]
    data: Data



from pathlib import Path
import tomllib
from typing import TypedDict


class Project(TypedDict):
    name: str
    version: str


def get_project(project: Path) -> Project:
    toml = project / "pyproject.toml"
    if not toml.exists():
        raise FileNotFoundError(f"Could not find {toml}")
    data = tomllib.loads(toml.read_text(encoding="utf-8"))
    return Project(data["project"])


def main():
    project = get_project(Path.cwd())
    version_path = Path("src") / project["name"] / "version.py"
    version_path.write_text(f"""\
# Do not edit this file, it is automatically generated by scripts/generate_version.py
VERSION = "{project['version']}"
""")


if __name__ == "__main__":
    main()



import re

import bs4

from iwashi.helper import HTTP_REGEX, normalize_url
from iwashi.visitor import Context, Service


class Booth(Service):
    def __init__(self) -> None:
        super().__init__(
            name="Booth",
            regex=re.compile(HTTP_REGEX + r"(?P<id>[\w-]+)\.booth\.pm", re.IGNORECASE),
        )

    async def visit(self, context: Context, id: str) -> None:
        url = f"https://{id}.booth.pm"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        name_element = soup.select_one(".shop-name")
        name = name_element.text if name_element is not None else None
        desc_element = soup.select_one(".booth-description .autolink")
        description = desc_element.text if desc_element is not None else None
        avater_element = soup.select_one(".avatar-image")

        context.create_result(
            self,
            id=id,
            url=url,
            name=name,
            description=description,
            profile_picture=avater_element.attrs["style"].split("url(")[1].split(")")[0]
            if avater_element is not None
            else None,
        )

        if desc_element is None:
            return
        for link in soup.select(".booth-description a"):
            if "href" not in link.attrs:
                continue
            normalized_url = normalize_url(link.attrs["href"])
            if normalized_url is None:
                continue
            context.enqueue_visit(normalized_url)



import pytest
from iwashi.service.litlink import LitLink
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_litlink():
    service = LitLink()
    correct = Result(
        service=service,
        id="leftpory5n",
        url="https://lit.link/leftpory5n",
        name="左ポリ５ン",
        description="ひだりぽりごんと読みます。",
        profile_picture="https://prd.storage.lit.link/images/creators/3f1f9feb-8603-46ff-a038-b2a29021beaa/687f1d68-9516-4e61-b38f-7a4377657687.png",
        links={
            "https://skeb.jp/@leftpory5n",
            "https://twitter.com/leftpory5n",
            "https://giftee.com/u/odenpo",
            "https://odaibako.net/u/leftpory5n",
            "https://www.pixiv.net/users/259841",
        },
    )
    await _test_service(service, correct, "https://lit.link/leftpory5n")



import pytest
from iwashi.service.twitter import Twitter
from iwashi.visitor import Result
from tests.service_tester import _test_service


@pytest.mark.asyncio
async def test_twitter():
    service = Twitter()
    correct = Result(
        service=service,
        id="VALIGHT1",
        url="https://twitter.com/VALIGHT1",
        name="VALIGHT_YT",
        description="Hi, I'm Valight, a creative Fortnite mappeur\nWelcome to my twiter account.\nCODE VALIGHT",
        profile_picture="https://pbs.twimg.com/profile_images/1512100269831639049/R4AjwP-9_normal.jpg",
        links={"https://youtube.com/channel/UCDRIfKm06-e1dwSbnU1H9Rw"},
    )
    await _test_service(service, correct, "https://twitter.com/VALIGHT1")



from __future__ import annotations

import json
import re
from typing import List, TypedDict

import bs4

from iwashi.helper import HTTP_REGEX
from iwashi.visitor import Context, Service

DATA_REGEX = r"preloadLink\s?=\s(?P<json>{[^;]+)"


class Bandcamp(Service):
    def __init__(self):
        super().__init__(
            name="Bandcamp",
            regex=re.compile(
                HTTP_REGEX + r"(?P<id>[\w-]+)?\.bandcamp\.com", re.IGNORECASE
            ),
        )

    async def visit(self, context: Context, id: str) -> None:
        url = f"https://{id}.bandcamp.com"
        res = await context.session.get(url)
        res.raise_for_status()
        soup = bs4.BeautifulSoup(await res.text(), "html.parser")
        script = soup.select_one("script[data-band]")
        if script is None:
            raise Exception("No script found")

        data: Root = json.loads(script.attrs["data-band"])
        description_element = soup.select_one(".signed-out-artists-bio-text")
        profile_picture_element = soup.select_one(".band-photo")
        context.create_result(
            self,
            id=id,
            url=url,
            name=data["name"],
            description=description_element.text.strip()
            if description_element
            else None,
            profile_picture=profile_picture_element.attrs["src"]
            if profile_picture_element
            else None,
        )

        for site in data["sites"]:
            context.enqueue_visit(site["url"])


class Site(TypedDict):
    url: str
    title: str


class Root(TypedDict):
    sites: List[Site]
    name: str
